/*
 * Kernel Module Wrapper
 * Copyright (C) 2001 Marco Lamberto <lm@sunnyspot.org>
 * Web page: http://kmw.sourceforge.net/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id$
 */

#ifndef _KMW_H_
#define _KMW_H_

#include <linux/config.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
#warning Linux Kernel 2.4.x: kmw procfs unsupported
#undef CONFIG_PROC_FS
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kmod.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#endif

#define KMW_VERSION					"0.1.1"
#define KMW_MAX_MODNAME			200
#define KMW_MAX_FNAME				200
#define KMW_MAX_MODULES			100					/* max number of sub modules */
#define KMW_MC_HOOK_SIZE		6						/* jmp opcode size - FIXME: i386 only */

#define TRUE								-1
#define FALSE								0


typedef void (*kmw_func)(void);

typedef struct HOOK {
	unsigned int	init : 1;										/* initialized */
	unsigned int	activate : 1;								/* activate this hook on loading */
	unsigned int	active : 1;									/* hook is active */
	char					name[KMW_MAX_FNAME + 1];		/* function name */
	kmw_func			f_org;											/* orginal kernel function */
	kmw_func			f_new;											/* replacement function */
	char 					b_org[KMW_MC_HOOK_SIZE];		/* original function entry point */
	char					b_new[KMW_MC_HOOK_SIZE];		/* jmp opcode */
#ifdef CONFIG_PROC_FS
	struct file_operations	file_ops;
	struct inode_operations	inode_ops;
	struct proc_dir_entry		file;							/* proc fs directory entry */
	struct HOOK 					**self;
#endif
} kmw_hook;

typedef int kmw_modid;											/* each wrapped module gets an id */


extern kmw_modid kmw_register_hooks(char *modname, kmw_hook *h_list);
extern int kmw_unregister_hooks(kmw_modid modid);
extern int kmw_toggle_hook(kmw_hook *hook, int status);
#ifdef CONFIG_PROC_FS
int module_open(struct inode *inode, struct file *file);
int module_close(struct inode *inode, struct file *file);
int module_permission(struct inode *inode, int op);
#endif


#define KMW_HOOK(func)\
	_kmw_hook_##func                                        
	                                                        
#define KMW_INIT(modname)\
	KMW_MODULE(modname)\
	kmw_hook kmw_hooks_##modname[] =\
		{                                                     
	                                                        
#ifndef CONFIG_PROC_FS
#define KMW_ADD_HOOK(func, on_load)\
			{\
				FALSE,\
				on_load,\
				FALSE,\
				#func,\
				(kmw_func)func,\
				(kmw_func)KMW_HOOK(func),\
				"",\
				"",\
			},
#else
#define KMW_ADD_HOOK(func, on_load)\
			{\
				FALSE,\
				on_load,\
				FALSE,\
				#func,\
				(kmw_func)func,\
				(kmw_func)KMW_HOOK(func),\
				"",\
				"",\
				{\
					NULL,\
					module_output_##func,\
					module_input_##func,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					module_open,\
					NULL,\
					module_close,\
				},\
				{\
					NULL, /* file ops goes here */\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					NULL,\
					module_permission\
				},\
				{\
					0,\
					1,			/* name length, this value is changed in kmw_init_hook */\
					#func,\
					S_IFREG | S_IWUSR | S_IRUGO,\
					1,\
					0, 0,\
					2,			/* file size (status char + \n) */\
					NULL,		/* inode ops goes here */\
					NULL\
				},\
				&hook_##func,\
			},
#endif
			                                                    
#ifndef CONFIG_PROC_FS
#define KMW_DONE()\
			{\
				FALSE,\
				FALSE,\
				FALSE,\
				"",\
				NULL,\
				NULL,\
				"",\
				"",\
			}\
		};
#else
#define KMW_DONE()\
			{\
				FALSE,\
				FALSE,\
				FALSE,\
				"",\
				NULL,\
				NULL,\
				"",\
				"",\
				{ NULL },\
				{ NULL },\
				{ NULL },\
				NULL\
			}\
		};
#endif

#define KMW_MODULE(modname)\
	extern kmw_hook kmw_hooks_##modname[];\
	kmw_modid kmw_modid_##modname = -1;\
\
	int init_module(void)\
	{\
		request_module("kmw");\
		kmw_modid_##modname = \
			kmw_register_hooks(#modname, kmw_hooks_##modname);\
		return kmw_modid_##modname < 0;\
	}\
\
	void cleanup_module(void)\
	{\
		kmw_unregister_hooks(kmw_modid_##modname);\
	}


#ifndef CONFIG_PROC_FS
#define KMW_PROC_ENTRY(func)\
	kmw_hook *hook_##func;\
\
	static int \
	module_output_##func(struct file *file, char *buf, size_t len,\
		loff_t *offset)\
	{\
		return 0;\
	}\
\
	static ssize_t \
	module_input_##func(struct file *file, char *buf, size_t len,\
		loff_t *offset)\
	{\
		return 1;\
	}

int
module_permission(struct inode *inode, int op)
{
	return -EACCES;
}

int
module_open(struct inode *inode, struct file *file)
{
	return 0;
}

int
module_close(struct inode *inode, struct file *file)
{
	return 0;
}
#else
#define KMW_PROC_ENTRY(func)\
	kmw_hook *hook_##func;\
\
	static int \
	module_output_##func(struct file *file, char *buf, size_t len,\
		loff_t *offset)\
	{\
		static char finished = 0;\
\
		if (finished) {\
			finished = 0;\
			return 0;\
		}\
		/*printk(KERN_INFO "KMW rstatus %s:%d\n",\
			hook_##func->name, hook_##func->active);*/\
		buf[0] = '0' + hook_##func->active;\
		buf[1] = '\n';\
		finished = 1;\
		return 2;\
	}\
\
	static ssize_t \
	module_input_##func(struct file *file, char *buf, size_t len,\
		loff_t *offset)\
	{\
		if (buf[0] == '0' || buf[0] == '1') {\
			/*printk(KERN_INFO "KMW wstatus %X\n", buf[0] - '0');*/\
			kmw_toggle_hook(hook_##func, buf[0] - '0');\
		}\
		return 1;\
	}

int
module_permission(struct inode *inode, int op)
{
	if (op == 4 || (op == 2 && current->euid == 0))
		return 0;

	return -EACCES;
}

int
module_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	return 0;
}

int
module_close(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}
#endif

#if __SMP__
#error ** NO SMP SUPPORT! **
#endif

#endif

/* vim: set ts=2: */
